"""
Includes modified code from [pymoo](https://github.com/anyoptimization/pymoo).

Sources:
    - [moead.py](https://github.com/anyoptimization/pymoo/blob/main/pymoo/algorithms/moo/moead.py).

Licensed under the Apache License, Version 2.0. Original copyright and license terms are preserved.

Add a method `_next_static_dynamic for dynamic multi-objective optimization.`
"""
import numpy as np
from pymoo.core.duplicate import NoDuplicateElimination
from pymoo.core.population import Population
from pymoo.core.selection import Selection
from pymoo.core.variable import Real, get
from pymoo.docs import parse_doc_string
from pymoo.operators.crossover.sbx import SBX
from pymoo.operators.mutation.pm import PM
from pymoo.operators.sampling.rnd import FloatRandomSampling
from pymoo.util.display.multi import MultiObjectiveOutput
from pymoo.util.reference_direction import default_ref_dirs
from scipy.spatial.distance import cdist

from pydmoo.algorithms.base.core.algorithm import LoopwiseAlgorithm
from pydmoo.algorithms.base.core.genetic import GeneticAlgorithm


class NeighborhoodSelection(Selection):

    def __init__(self, prob=1.0) -> None:
        super().__init__()
        self.prob = Real(prob, bounds=(0.0, 1.0))

    def _do(self, problem, pop, n_select, n_parents, neighbors=None, random_state=None, **kwargs):
        assert n_select == len(neighbors)
        P = np.full((n_select, n_parents), -1)

        prob = get(self.prob, size=n_select)

        for k in range(n_select):
            if random_state.random() < prob[k]:
                P[k] = random_state.choice(neighbors[k], n_parents, replace=False)
            else:
                P[k] = random_state.permutation(len(pop))[:n_parents]

        return P


# =========================================================================================================
# Implementation
# =========================================================================================================

class MOEAD(LoopwiseAlgorithm, GeneticAlgorithm):

    def __init__(self,
                 ref_dirs=None,
                 n_neighbors=20,
                 decomposition=None,
                 prob_neighbor_mating=0.9,
                 sampling=FloatRandomSampling(),
                 crossover=SBX(prob=1.0, eta=20),
                 mutation=PM(prob_var=None, eta=20),
                 output=MultiObjectiveOutput(),
                 **kwargs):

        # reference directions used for MOEAD
        self.ref_dirs = ref_dirs

        # the decomposition metric used
        self.decomposition = decomposition

        # the number of neighbors considered during mating
        self.n_neighbors = n_neighbors

        self.neighbors = None

        self.selection = NeighborhoodSelection(prob=prob_neighbor_mating)

        super().__init__(pop_size=len(ref_dirs),
                         sampling=sampling,
                         crossover=crossover,
                         mutation=mutation,
                         eliminate_duplicates=NoDuplicateElimination(),
                         output=output,
                         advance_after_initialization=False,
                         **kwargs)

    def _setup(self, problem, **kwargs):
        assert not problem.has_constraints(), "This implementation of MOEAD does not support any constraints."

        # if no reference directions have been provided get them and override the population size and other settings
        if self.ref_dirs is None:
            self.ref_dirs = default_ref_dirs(problem.n_obj)
        self.pop_size = len(self.ref_dirs)

        # neighbours includes the entry by itself intentionally for the survival method
        self.neighbors = np.argsort(cdist(self.ref_dirs, self.ref_dirs), axis=1, kind='quicksort')[:, :self.n_neighbors]

        # if the decomposition is not set yet, set the default
        if self.decomposition is None:
            self.decomposition = default_decomp(problem)

    def _initialize_advance(self, infills=None, **kwargs):
        super()._initialize_advance(infills, **kwargs)
        self.ideal = np.min(self.pop.get("F"), axis=0)

    def _next(self):
        # Added by DynOpt Team on Dec 21, 2025
        pop = self._next_static_dynamic()

        # iterate for each member of the population in random order
        for k in self.random_state.permutation(len(pop)):
            # get the parents using the neighborhood selection
            P = self.selection.do(self.problem, pop, 1, self.mating.crossover.n_parents, neighbors=[self.neighbors[k]], random_state=self.random_state)

            # perform a mating using the default operators - if more than one offspring just pick the first
            off = self.random_state.choice(self.mating.do(self.problem, pop, 1, parents=P, n_max_iterations=1, random_state=self.random_state))

            # evaluate the offspring
            off = yield off

            # update the ideal point
            self.ideal = np.min(np.vstack([self.ideal, off.F]), axis=0)

            # now actually do the replacement of the individual is better
            self._replace(k, off)

    # Added by DynOpt Team on Dec 21, 2025
    def _next_static_dynamic(self):
        pop = self.pop

        return pop

    def _replace(self, k, off):
        pop = self.pop

        # calculate the decomposed values for each neighbor
        N = self.neighbors[k]
        FV = self.decomposition.do(pop[N].get("F"), weights=self.ref_dirs[N, :], ideal_point=self.ideal)
        off_FV = self.decomposition.do(off.F[None, :], weights=self.ref_dirs[N, :], ideal_point=self.ideal)

        # this makes the algorithm to support constraints - not originally proposed though and not tested enough
        # if self.problem.has_constraints():
        #     CV, off_CV = pop[N].get("CV")[:, 0], np.full(len(off_FV), off.CV)
        #     fmax = max(FV.max(), off_FV.max())
        #     FV, off_FV = parameter_less(FV, CV, fmax=fmax), parameter_less(off_FV, off_CV, fmax=fmax)

        # get the absolute index in F where offspring is better than the current F (decomposed space)
        I = np.where(off_FV < FV)[0]
        pop[N[I]] = off


class ParallelMOEAD(MOEAD):

    def __init__(self, ref_dirs, **kwargs):
        super().__init__(ref_dirs, **kwargs)
        self.indices = None

    def _infill(self):
        pop_size, cross_parents, cross_off = self.pop_size, self.mating.crossover.n_parents, self.mating.crossover.n_offsprings

        # do the mating in a random order
        indices = self.random_state.permutation(len(self.pop))[:self.n_offsprings]

        # get the parents using the neighborhood selection
        P = self.selection.do(self.problem, self.pop, self.n_offsprings, cross_parents,
                              neighbors=self.neighbors[indices], random_state=self.random_state)

        # do not any duplicates elimination - thus this results in exactly pop_size * n_offsprings offsprings
        off = self.mating.do(self.problem, self.pop, 1e12, n_max_iterations=1, parents=P, random_state=self.random_state)

        # select a random offspring from each mating
        off = Population.create(*[self.random_state.choice(pool) for pool in np.reshape(off, (self.n_offsprings, -1))])

        # store the indices because of the neighborhood matching in advance
        self.indices = indices

        return off

    def _advance(self, infills=None, **kwargs):
        assert len(self.indices) == len(infills), "Number of infills must be equal to the one created beforehand."

        # update the ideal point before starting to replace
        self.ideal = np.min(np.vstack([self.ideal, infills.get("F")]), axis=0)

        # now do the replacements as in the loop-wise version
        for k, off in enumerate(infills):
            self._replace(self.indices[k], off)


def default_decomp(problem):
    if problem.n_obj <= 2:
        from pymoo.decomposition.tchebicheff import Tchebicheff
        return Tchebicheff()
    else:
        from pymoo.decomposition.pbi import PBI
        return PBI()


parse_doc_string(MOEAD.__init__)



import numpy as np


def manifold_prediction(X0, X1):
    """
    Predict the next manifold and compute its dispersion metric.

    Given two sequential populations in a manifold, this method:
    1. Centers the populations by removing their means
    2. Computes a dispersion metric (sigma) based on set distances

    Parameters
    ----------
    X0 : ndarray, shape (N, n)
        Population at time t-1, where:
        - N: number of points/samples
        - n: dimensionality of the manifold
    X1 : ndarray, shape (N, n)
        Population at time t

    Returns
    -------
    tuple (ndarray, float)
        - C1: ndarray, shape (N, n)
            Centered manifold at time t
        - variance: float
            Normalized dispersion metric computed as:
            variance = (D(C1,C0)^2) / n
            where D is the set distance between manifolds

    Notes
    -----
    1. The manifold is constructed by centering the input data
    2. The dispersion metric variance quantifies the normalized squared distance
       between consecutive manifolds
    3. Requires self.set_distance() method to be implemented
    4. Both input arrays must have same shape (N, n)

    """
    # Center the populations by removing column means
    C0 = X0 - np.mean(X0, axis=0)  # Centered manifold at t-1
    C1 = X1 - np.mean(X1, axis=0)  # Centered manifold at t

    return C1, set_distance(C1, C0)


def set_distance(A, B):
    """
    Compute the average minimum distance between two sets of points.

    The distance metric is defined as:
    D(A, B) = (1/|A|) ∑_{x∈A} min_{y∈B} ||x - y||

    Parameters
    ----------
    A : ndarray, shape (m, d)
        First set of points containing m samples in d-dimensional space.
    B : ndarray, shape (n, d)
        Second set of points containing n samples in d-dimensional space.

    Returns
    -------
    float
        The average minimum Euclidean distance between points in A and their
        nearest neighbors in B.

    Notes
    -----
    1. Uses Euclidean (L2) norm for distance computation.
    2. For empty sets, will raise ValueError.
    3. Computational complexity is O(m*n*d) where:
       - m = number of points in A
       - n = number of points in B
       - d = dimensionality

    Examples
    --------
    >>> A = np.array([[1.0, 2.0], [3.0, 4.0]])
    >>> B = np.array([[1.1, 2.1], [3.1, 4.1], [5.0, 6.0]])
    >>> set_distance(A, B)
    0.14142135623730953
    """
    # Compute pairwise Euclidean distances using broadcasting:
    # A[:, np.newaxis, :] reshapes to (m, 1, d)
    # B[np.newaxis, :, :] reshapes to (1, n, d)
    # Resulting subtraction produces (m, n, d) array
    distances = np.linalg.norm(A[:, np.newaxis, :] - B[np.newaxis, :, :], axis=2)

    # Find minimum distance for each point in A to any point in B
    min_distances = np.min(distances, axis=1)

    # Return average minimum distance
    return np.mean(min_distances)



import numpy as np
from scipy.linalg import eigh


class TCA:
    """Transfer Component Analysis (TCA) for domain adaptation.

    TCA finds a latent space where source and target domain distributions are similar.

    Parameters
    ----------
    kernel_type : {'linear', 'rbf', 'poly'}, default='rbf'
        Type of kernel function to use:
        - 'linear': Linear kernel (dot product)
        - 'rbf': Radial Basis Function (Gaussian) kernel
        - 'poly': Polynomial kernel
    kernel_param : float, default=1.0
        Parameter for kernel function:
        - For 'rbf': gamma parameter (inverse of kernel width)
        - For 'poly': degree of polynomial
    dim : int, default=20
        Dimensionality of the latent space (number of components to keep)
    mu : float, default=0.5
        Regularization parameter for numerical stability

    Attributes
    ----------
    W : ndarray of shape (n_samples, dim) or None
        Learned transformation matrix. None before fitting.
    X_train : ndarray of shape (n_samples, n_features) or None
        Training data stored for transformation. None before fitting.
    """

    def __init__(self, kernel_type='rbf', kernel_param=1.0, dim=20, mu=0.5):
        # For the TCA parameters, we set the Gaussian kernel function to the default value and the expected dimensionality
        # was set to be 20. The value of μ was set to 0.5.
        self.kernel_type = kernel_type
        self.kernel_param = kernel_param
        self.dim = dim
        self.mu = mu
        self.W = None
        self.X_train = None

    def fit(self, Xs, Xt):
        """Fit TCA model to source and target data.

        Parameters
        ----------
        Xs : ndarray of shape (ns_samples, n_features)
            Source domain feature matrix
        Xt : ndarray of shape (nt_samples, n_features)
            Target domain feature matrix

        Returns
        -------
        self : object
            Returns the instance itself.
        """
        # Stack source and target data vertically
        X = np.vstack((Xs, Xt))
        n = X.shape[0]  # Total number of samples
        ns, nt = Xs.shape[0], Xt.shape[0]  # Number of source/target samples

        # Compute kernel matrix K using selected kernel function
        K = self._kernel(X, X)
        self.X_train = X  # Store for transform phase

        # Construct MMD (Maximum Mean Discrepancy) matrix
        # This matrix encodes the distribution difference between domains
        M = np.zeros((n, n))
        M[:ns, :ns] = 1/(ns*ns)  # Source-source block
        M[ns:, ns:] = 1/(nt*nt)  # Target-target block
        M[:ns, ns:] = M[ns:, :ns] = -1/(ns*nt)  # Cross blocks

        # Centering matrix H = I - 1/n * 11^T
        # Projects data onto the space orthogonal to the vector of ones
        H = np.eye(n) - np.ones((n, n))/n

        # Solve generalized eigenvalue problem:
        # (K(M)K + μI)^-1 K(H)K w = λw
        # Using pseudo-inverse for numerical stability
        A = np.linalg.pinv(K @ M @ K + self.mu*np.eye(n)) @ K @ H @ K

        # Compute eigenvalues and eigenvectors
        # eigh returns eigenvalues in ascending order
        eigvals, eigvecs = eigh(A)

        # Select top 'dim' eigenvectors corresponding to largest eigenvalues
        # [::-1] reverses order to get descending eigenvalues
        self.W = eigvecs[:, np.argsort(eigvals)[-self.dim:][::-1]]

        return self

    def transform(self, X):
        """Transform data using learned TCA components.

        Parameters
        ----------
        X : ndarray of shape (n_samples, n_features)
            Data to transform

        Returns
        -------
        X_transformed : ndarray of shape (n_samples, dim)
            Data projected to the latent space

        Raises
        ------
        ValueError
            If fit() hasn't been called before transform()
        """
        if self.W is None:
            raise ValueError("Model not fitted yet. Call fit() first.")

        # Compute kernel between new data and training data
        K = self._kernel(X, self.X_train)

        # Project to latent space: X' = K * W
        return K @ self.W

    def _kernel(self, X1, X2):
        """Compute kernel matrix between X1 and X2.

        Parameters
        ----------
        X1 : ndarray of shape (n_samples1, n_features)
            First set of samples
        X2 : ndarray of shape (n_samples2, n_features)
            Second set of samples

        Returns
        -------
        K : ndarray of shape (n_samples1, n_samples2)
            Kernel matrix

        Raises
        ------
        ValueError
            If kernel_type is not supported
        """
        if self.kernel_type == 'linear':
            return X1 @ X2.T  # Linear kernel: K(x,y) = x^T y
        elif self.kernel_type == 'rbf':
            # RBF kernel: K(x,y) = exp(-\gamma||x-y||^2)
            # Efficient computation using ||x-y||^2 = ||x||^2 + ||y||^2 - 2x^T y
            dist_sq = (np.sum(X1**2, axis=1)[:, np.newaxis] + np.sum(X2**2, axis=1) - 2 * X1 @ X2.T)
            return np.exp(-self.kernel_param * dist_sq)
        elif self.kernel_type == 'poly':
            # Polynomial kernel: K(x,y) = (x^T y + 1)^d
            return (X1 @ X2.T + 1) ** self.kernel_param
        else:
            raise ValueError(f"Unsupported kernel type: {self.kernel_type}. Choose from 'linear', 'rbf', 'poly'")



"""
Includes modified code from [pymoo](https://github.com/anyoptimization/pymoo).

> Sources:
>
> - [dyn.py](https://github.com/anyoptimization/pymoo/blob/main/pymoo/problems/dyn.py)
>
> Licensed under the Apache License, Version 2.0. Original copyright and license terms are preserved.
"""

from abc import ABC
from math import ceil

from mpmath import mp
from pymoo.core.callback import Callback
from pymoo.core.problem import Problem


class DynamicProblem(Problem, ABC):
    """Abstract base class for dynamic optimization problems."""
    pass


class DynamicApplProblem(DynamicProblem):
    """Dynamic optimization problem for real-world applications.

    This class defines dynamic optimization problems that model practical, real-world scenarios where the problem
    characteristics change systematically over time.

    Parameters
    ----------
    nt : int
        Severity of change. Controls how significantly the problem changes
        at each change point. Higher values indicate more substantial changes
        in problem characteristics.
    taut : int
        Frequency of change. Specifies how often (in generations) the problem
        undergoes changes. Lower values mean more frequent changes.
    t0 : int, optional
        The first change occurs after t0 generations, by default 50.
        That is, the generation at which a change occurs is (t0+1), (t0+taut+1), etc.
        This allows for an initial stabilization period before the first change.
    tau : int, optional
        Current simulation time counter (in generations), by default 1.
    time : float, optional
        Explicit simulation time value (overrides calculated time), by default None.
        Used for manual time control in specific scenarios.
    **kwargs : dict
        Additional keyword arguments passed to the parent Problem class.

    Attributes
    ----------
    tau : int
        Current simulation time counter in generations.
    nt : int
        Severity of change at each change point.
    taut : int
        Frequency of change between consecutive changes.
    t0 : int
        Initial stabilization period before first change occurs.

    Notes
    -----
    This class models real-world dynamic scenarios where:

    - Changes occur at predictable intervals (every `taut` generations)
    - Change severity is controlled by `nt` parameter
    - Initial period `t0` allows for system stabilization
    """

    def __init__(self, nt: int, taut: int, t0: int = 50, tau: int = 1, time: float | None = None, **kwargs):
        super().__init__(**kwargs)
        self.tau = tau  # time counter
        self.nt = nt  # severity of change
        self.taut = taut  # frequency of change
        self.t0 = t0  # Initial time offset
        self._time = time

    def tic(self, elapsed: int = 1) -> None:

        # increase the time counter by one
        self.tau += elapsed

        # remove the cache of the problem to recreate ps and pf
        self.__dict__["cache"] = {}

    @property
    def time(self) -> float:
        if self._time is not None:
            return self._time
        else:
            # return 1 / self.nt * (self.tau // self.taut)

            # Calculate base time step
            delta_time = 1 / self.nt

            # Calculate time count considering initial offset
            count = max((self.tau + self.taut - (self.t0 + 1)), 0) // self.taut

            # Return time value
            return delta_time * count

    @time.setter
    def time(self, value: float) -> None:
        self._time = value

    def update_to_next_time(self):
        """Advance problem to the next significant time step.

        Returns
        -------
            elapsed: The actual time units advanced
        """
        # Calculate how many time steps to advance
        count = max((self.tau + self.taut - (self.t0 + 1)), 0) // self.taut

        # Calculate exact elapsed time needed to reach next discrete time point
        elapsed = int(count * self.taut + (self.t0 + 1) - self.tau)

        # Advance time by calculated amount
        self.tic(elapsed=elapsed)

        return elapsed


class DynamicTestProblem(DynamicProblem):
    """Dynamic optimization problem for testing and benchmarking.

    Parameters
    ----------
    nt : int
        Severity of change. Controls how significantly the problem changes
        at each change point. Higher values indicate more substantial changes
        in problem characteristics.
    taut : int
        Frequency of change. Specifies how often (in generations) the problem
        undergoes changes. Lower values mean more frequent changes.
    t0 : int, optional
        The first change occurs after t0 generations, by default 50.
        That is, the generation at which a change occurs is (t0+1), (t0+taut+1), etc.
        This allows for an initial stabilization period before the first change.
    tau : int, optional
        Current simulation time counter (in generations), by default 1.
    time : float, optional
        Explicit simulation time value (overrides calculated time), by default None.
        Used for manual time control in specific scenarios.
    add_time_perturbation : bool, optional
        If True, adds perturbations to the time calculation, by default False.
    **kwargs : dict
        Additional keyword arguments passed to the parent Problem class.

    Attributes
    ----------
    tau : int
        Current simulation time counter in generations.
    nt : int
        Severity of change at each change point.
    taut : int
        Frequency of change between consecutive changes.
    t0 : int
        Initial stabilization period before first change occurs.
    add_time_perturbation : bool
        Flag indicating whether to add stochastic perturbations.

    Notes
    -----
    This class is designed for testing scenarios where:

    - Changes occur at predictable intervals (every `taut` generations)
    - Change severity is controlled by `nt` parameter
    - Initial period `t0` allows for system stabilization
    - Stochastic perturbations can be added for more complex testing
    - Reproducibility is important for benchmarking
    """

    def __init__(self, nt, taut, t0=50, tau=1, time=None, add_time_perturbation=False, **kwargs):
        super().__init__(**kwargs)
        self.tau = tau  # time counter
        self.nt = nt  # severity of change
        self.taut = taut  # frequency of change
        self.t0 = t0  # Initial time offset - added by DynOpt Team
        self._time = time

        self.add_time_perturbation = add_time_perturbation  # Stochastic perturbation flag - added by DynOpt Team

    def tic(self, elapsed=1):

        # increase the time counter by one
        self.tau += elapsed

        # remove the cache of the problem to recreate ps and pf
        self.__dict__["cache"] = {}

    @property
    def time(self):
        if self._time is not None:
            return self._time
        else:
            # return 1 / self.nt * (self.tau // self.taut)

            # Modified by DynOpt Team
            # Calculate base time step
            delta_time = 1 / self.nt

            # Calculate time count considering initial offset
            count = max((self.tau + self.taut - (self.t0 + 1)), 0) // self.taut

            # Calculate perturbation ratio if enabled
            if not self.add_time_perturbation:
                ratio = 0

            else:
                # Use mathematical constants to generate deterministic perturbations
                mp.dps = max(ceil(10 + count), 10)
                mp_pi = 0 if count == 0 else int(str(mp.pi).split(".")[-1][count - 1])  # Extract digit from pi
                ratio = 0.5 * 1 / 9 * mp_pi

            # Return time value with optional perturbation
            return delta_time * count + delta_time * ratio

    @time.setter
    def time(self, value):
        self._time = value

    # Added by DynOpt Team
    def update_to_next_time(self):
        """Advance problem to the next significant time step.

        Returns
        -------
            elapsed: The actual time units advanced
        """
        # Calculate how many time steps to advance
        count = max((self.tau + self.taut - (self.t0 + 1)), 0) // self.taut

        # Calculate exact elapsed time needed to reach next discrete time point
        elapsed = int(count * self.taut + (self.t0 + 1) - self.tau)

        # Advance time by calculated amount
        self.tic(elapsed=elapsed)

        return elapsed


class TimeSimulation(Callback):
    """Callback for simulating time evolution in dynamic optimization problems.

    Handles time-linkage properties and time step updates.
    """

    def update(self, algorithm):
        """Update method called at each algorithm iteration."""
        problem = algorithm.problem

        # Added by DynOpt Team
        # Emulate time-linkage property: Update problem state based on current optimal solutions
        # Must execute before the problem.tic() to ensure proper time sequencing
        if hasattr(problem, "time_linkage") and hasattr(problem, "cal"):
            # Calculate time-linkage effects using current optimal objective values
            problem.cal(algorithm.opt.get("F"))

        # Advance time step for dynamic problem simulation
        if hasattr(problem, "tic"):
            problem.tic()  # Progress the dynamic problem to next time step
        else:
            raise Exception("TimeSimulation can only be used for dynamic test problems.")



"""
Includes modified code from [pymoo](https://github.com/anyoptimization/pymoo).

Sources:
    - [nsga2.py](https://github.com/anyoptimization/pymoo/blob/main/pymoo/algorithms/moo/nsga2.py).

Licensed under the Apache License, Version 2.0. Original copyright and license terms are preserved.
"""

import warnings

import numpy as np
from pymoo.docs import parse_doc_string
from pymoo.operators.crossover.sbx import SBX
from pymoo.operators.mutation.pm import PM
from pymoo.operators.sampling.rnd import FloatRandomSampling
from pymoo.operators.selection.tournament import TournamentSelection, compare
from pymoo.operators.survival.rank_and_crowding import RankAndCrowding
from pymoo.termination.default import DefaultMultiObjectiveTermination
from pymoo.util.display.multi import MultiObjectiveOutput
from pymoo.util.dominator import Dominator
from pymoo.util.misc import has_feasible

from pydmoo.algorithms.base.core.genetic import GeneticAlgorithm

# ---------------------------------------------------------------------------------------------------------
# Binary Tournament Selection Function
# ---------------------------------------------------------------------------------------------------------


def binary_tournament(pop, P, algorithm, **kwargs):
    n_tournaments, n_parents = P.shape

    if n_parents != 2:
        raise ValueError("Only implemented for binary tournament!")

    tournament_type = algorithm.tournament_type
    S = np.full(n_tournaments, np.nan)

    for i in range(n_tournaments):

        a, b = P[i, 0], P[i, 1]
        a_cv, a_f, b_cv, b_f = pop[a].CV[0], pop[a].F, pop[b].CV[0], pop[b].F
        rank_a, cd_a = pop[a].get("rank", "crowding")
        rank_b, cd_b = pop[b].get("rank", "crowding")

        # if at least one solution is infeasible
        if a_cv > 0.0 or b_cv > 0.0:
            S[i] = compare(a, a_cv, b, b_cv, method='smaller_is_better', return_random_if_equal=True, random_state=algorithm.random_state)

        # both solutions are feasible
        else:

            if tournament_type == 'comp_by_dom_and_crowding':
                rel = Dominator.get_relation(a_f, b_f)
                if rel == 1:
                    S[i] = a
                elif rel == -1:
                    S[i] = b

            elif tournament_type == 'comp_by_rank_and_crowding':
                S[i] = compare(a, rank_a, b, rank_b, method='smaller_is_better')

            else:
                raise Exception("Unknown tournament type.")

            # if rank or domination relation didn't make a decision compare by crowding
            if np.isnan(S[i]):
                S[i] = compare(a, cd_a, b, cd_b, method='larger_is_better', return_random_if_equal=True, random_state=algorithm.random_state)

    return S[:, None].astype(int, copy=False)


# ---------------------------------------------------------------------------------------------------------
# Survival Selection
# ---------------------------------------------------------------------------------------------------------


class RankAndCrowdingSurvival(RankAndCrowding):

    def __init__(self, nds=None, crowding_func="cd"):
        super().__init__(nds, crowding_func)

# =========================================================================================================
# Implementation
# =========================================================================================================


class NSGA2(GeneticAlgorithm):

    def __init__(self,
                 pop_size=100,
                 sampling=FloatRandomSampling(),
                 selection=TournamentSelection(func_comp=binary_tournament),
                 crossover=SBX(eta=15, prob=0.9),
                 mutation=PM(eta=20),
                 survival=RankAndCrowding(),
                 output=MultiObjectiveOutput(),
                 **kwargs):

        super().__init__(
            pop_size=pop_size,
            sampling=sampling,
            selection=selection,
            crossover=crossover,
            mutation=mutation,
            survival=survival,
            output=output,
            advance_after_initial_infill=True,
            **kwargs)

        self.termination = DefaultMultiObjectiveTermination()
        self.tournament_type = 'comp_by_dom_and_crowding'

    def _set_optimum(self, **kwargs):
        if not has_feasible(self.pop):
            self.opt = self.pop[[np.argmin(self.pop.get("CV"))]]
        else:
            self.opt = self.pop[self.pop.get("rank") == 0]


parse_doc_string(NSGA2.__init__)



import numpy as np
from scipy.linalg import logm, sqrtm


def wasserstein_distance(mean1, cov1, mean2, cov2):
    diff = mean1 - mean2
    sqrt_cov1 = sqrtm(cov1)
    cross_term = sqrtm(sqrt_cov1 @ cov2 @ sqrt_cov1)
    return np.sqrt(diff @ diff + np.trace(cov1 + cov2 - 2 * cross_term))


def kl_div_distance(mean1, cov1, mean2, cov2):
    inv_cov2 = np.linalg.inv(cov2)
    diff = mean2 - mean1
    logdet = np.log(np.linalg.det(cov2) / np.linalg.det(cov1))
    return 0.5 * (np.trace(inv_cov2 @ cov1) + diff.T @ inv_cov2 @ diff - len(mean1) + logdet)


def bhattacharyya_distance(mean1, cov1, mean2, cov2):
    avg_cov = 0.5 * (cov1 + cov2)
    inv_avg = np.linalg.inv(avg_cov)
    diff = mean1 - mean2
    return 0.125 * diff.T @ inv_avg @ diff + 0.5 * np.log(
        np.linalg.det(avg_cov) / np.sqrt(np.linalg.det(cov1) * np.linalg.det(cov2))
    )


def mahalanobis_distance(mean1, cov1, mean2, cov2):
    inv_cov = np.linalg.inv(0.5 * (cov1 + cov2))
    return np.sqrt((mean1 - mean2).T @ inv_cov @ (mean1 - mean2))


def riemannian_distance(cov1, cov2):
    return np.linalg.norm(logm(np.linalg.inv(sqrtm(cov1)) @ cov2 @ np.linalg.inv(sqrtm(cov1))), "fro")


def frobenius_distance(cov1, cov2):
    return np.linalg.norm(cov1 - cov2, "fro")


# Used in NSGA2-IMKT-N
# def norm_mean_frobenius_distance(mean, cov, target_mean, target_cov):
#     return np.linalg.norm(mean - target_mean) + frobenius_distance(cov, target_cov)
def norm_mean_frobenius_distance(mean, cov, target_mean, target_cov):
    return frobenius_distance(cov, target_cov)



import numpy as np


def uniform_sample_in_neighborhood(x, radius=0.1, bounds=None, size=1, random_state=None):
    """
    Generate uniformly distributed samples within hypercube neighborhoods of input points.

    Parameters
    ----------
    x : ndarray, shape (m, n)
        Input matrix of m solutions in n-dimensional space.
    radius : float or array_like, optional
        Neighborhood sampling radius. Can be:
        - Scalar: uniform radius for all dimensions (default=0.1)
        - Array of shape (n,): dimension-specific radii
        - Array of shape (m, n): solution-specific and dimension-specific radii
    bounds : tuple of array_like, optional
        Tuple containing (lower_bound, upper_bound) where each is of shape (n,).
        If provided, samples will be clipped to these bounds.
    size : int, optional
        Number of samples to generate per solution (default=1).

    Returns
    -------
    ndarray
        Sampled solutions:
        - If size=1: shape (m, n)
        - If size>1: shape (m, size, n)

    Notes
    -----
    1. Sampling is performed using uniform distribution within [-radius, +radius]
       around each solution coordinate.
    2. For bounded sampling, uses np.clip to enforce constraints.
    3. The function preserves input dtype for numerical precision.

    Examples
    --------
    >>> x = np.array([[1.0, 2.0], [3.0, 4.0]])
    >>> # Single sample with uniform radius
    >>> samples = uniform_sample_in_neighborhood(x, radius=0.5)
    >>> # Multiple samples with dimension-specific radii
    >>> samples = uniform_sample_in_neighborhood(x, radius=[0.1, 0.2], size=10)
    """
    # Input validation and shape processing
    x = np.asarray(x)
    m, n = x.shape  # m solutions, n dimensions

    # Process radius parameter
    if np.isscalar(radius):
        radius = np.full((1, n), radius)  # Broadcast scalar to all dimensions
    elif isinstance(radius, (list, np.ndarray)) and len(radius) == n:
        radius = np.reshape(radius, (1, n))  # Convert 1D vector to row vector

    # Generate random perturbations
    samples = x[:, np.newaxis, :] + random_state.uniform(low=-radius, high=radius, size=(m, size, n))  # np.random

    # Apply bounds constraint if provided
    if bounds is not None:
        lb, ub = np.asarray(bounds[0]), np.asarray(bounds[1])
        samples = np.clip(samples, lb, ub)

    return samples.squeeze()



import numpy as np
from pymoo.core.population import Population
from pymoo.decomposition.tchebicheff import Tchebicheff
from pymoo.operators.mutation.pm import PM

from pydmoo.algorithms.base.moo.moead import MOEAD


class MOEADDE(MOEAD):
    """MOEA/D-DE (Updated by Cao).

    It is worth noting that there is a distinct modification in line 28 compared with the original framework of MOEA/D-DE.
    The newly generated solution competes with each member from the corresponding mating neighborhood (denoted as Pool in Algorithm 2).
    But in the original MOEA/D-DE framework, it only competes with two members from the corresponding mating neighborhood.
    This modification expands the replacement neighborhood to enhance the exploitation capability that is extremely important in dealing with DMOPs.

    Cao, L., Xu, L., Goodman, E. D., Bao, C., and Zhu, S. (2020).
    Evolutionary dynamic multiobjective optimization assisted by a support vector regression predictor.
    IEEE Transactions on Evolutionary Computation, 24(2), 305–319.
    https://doi.org/10.1109/TEVC.2019.2925722

    Hui Li and Qingfu Zhang. (2009).
    Multiobjective optimization problems with complicated pareto sets, MOEA/D and NSGA-II.
    IEEE Transactions on Evolutionary Computation, 13(2), 284–302.
    https://doi.org/10.1109/TEVC.2008.925798
    """
    def __init__(self, decomposition=Tchebicheff(), prob_neighbor_mating=0.8, **kwargs):
        super().__init__(decomposition=decomposition, prob_neighbor_mating=prob_neighbor_mating, **kwargs)

        # Neighborhood Selection
        # the number of neighbors considered during mating
        self.n_neighbors = 20
        self._delta = 0.8  # 0.9 (prob_neighbor_mating)

        # DE crossover
        self._cr = 0.5  # 1
        self._F_ = 0.5  # 0.5

        # Polynomial mutation
        self._eta = 20
        # self._pm = 1/d  # d is the number of variables
        self.mutation = PM(eta=self._eta)  # prob_var=1/d, default is min(0.5, 1/d)

    def _next(self):
        pop = self._next_static_dynamic()

        # iterate for each member of the population in random order
        for k in self.random_state.permutation(len(pop)):
            # Step 2.1 Selection of Mating/Update Range:
            # Select parents - use neighborhood with probability delta, else global selection
            pp = self.neighbors[k] if self.random_state.random() < self._delta else list(range(self.pop_size))

            # Step 2.2 Reproduction
            # Randomly select three distinct parents from the chosen pool
            a, b, c = self.random_state.choice(pp, size=3, replace=True)

            # Initialize problem parameters
            n_var = self.problem.n_var
            xl, xu = self.problem.xl, self.problem.xu

            X = pop.get("X")  # Position critical

            # mutation operator
            # Differential evolution mutation: v = x1 + F*(x2 - x3)
            mutation = X[a] + self._F_ * (X[b] - X[c])

            # crossover operator
            # Create mask for crossover operations (CR probability)
            mask = self.random_state.random(n_var) < self._cr

            # Combine mutation with target vector based on crossover probability
            V = np.where(mask, mutation, X[k])

            # mutation operator
            # polynomial mutation operator
            r = self.random_state.random(n_var)
            delta = np.where(
                self.random_state.random(n_var) < 0.5,
                np.power(2*r, 1/(1+self._eta)) - 1,  # First perturbation formula
                1 - np.power(2-2*r, 1/(1+self._eta))  # Alternative perturbation formula
            )
            perturb_mask = self.random_state.random(n_var) < (1/n_var)  # self._pm
            V[perturb_mask] += delta[perturb_mask] * (xu[perturb_mask] - xl[perturb_mask])

            # Step 2.3 Repair: If an element is out of the boundary, its value is reset to be a randomly selected value inside the boundary.
            # Ensure solution stays within bounds
            V = np.clip(V, xl, xu)

            # Individual
            off = Population.new(X=np.array([V]))[0]

            # evaluate the offspring
            off = yield off

            # Step 2.4 Update
            # update the ideal point
            self.ideal = np.min(np.vstack([self.ideal, off.F]), axis=0)

            # Step 2.5 Update of Solutions
            # It is worth noting that there is a distinct modification in line 28 compared with the original framework of MOEA/D-DE.
            # The newly generated solution competes with each member from the corresponding mating neighborhood (denoted as Pool in Algorithm 2).
            # But in the original MOEA/D-DE framework, it only competes with two members from the corresponding mating neighborhood.
            # This modification expands the replacement neighborhood to enhance the exploitation capability that is extremely important in dealing with DMOPs.
            # now actually do the replacement of the individual is better
            self._replace(k, off)



import numpy as np
from pymoo.core.population import Population
from pymoo.operators.survival.rank_and_crowding import RankAndCrowding

from pydmoo.algorithms.base.dmoo.dmoeadde import DMOEADDE
from pydmoo.core.inverse import closed_form_solution


class MOEADDEAE(DMOEADDE):
    """Autoencoding.

    References
    ----------
    Feng, L., Zhou, W., Liu, W., Ong, Y.-S., and Tan, K. C. (2022).
    Solving dynamic multiobjective problem via autoencoding evolutionary search.
    IEEE Transactions on Cybernetics, 52(5), 2649–2662.
    https://doi.org/10.1109/TCYB.2020.3017017
    """

    def __init__(self, **kwargs):

        super().__init__(**kwargs)

    def _response_mechanism(self):
        """Response mechanism."""
        pop = self.pop
        X = pop.get("X")

        # recreate the current population without being evaluated
        pop = Population.new(X=X)

        # predict via denoising autoencoding
        PSs = self.data.get("PSs", [])
        PSs.append(self.opt.get("X"))  # Parate Set
        PSs = PSs[-2:]
        self.data["PSs"] = PSs

        a = 0
        if len(PSs) == 2:
            # Pareto Set
            P, Q = PSs

            # Q = PM
            min_len = min(len(P), len(Q))
            M = closed_form_solution(Q[:min_len], P[:min_len])

            # X = QM
            X = np.dot(Q, M)

            # bounds
            if self.problem.has_bounds():
                xl, xu = self.problem.bounds()
                X = np.clip(X, xl, xu)  # not provided in the original reference literature

            # evalutate new population
            samples = self.evaluator.eval(self.problem, Population.new(X=X))
            a = min(int(self.pop_size / 2), len(samples))

            # do a survival to recreate rank and crowding of all individuals
            samples = RankAndCrowding().do(self.problem, samples, n_survive=a, random_state=self.random_state)

            pop[:a] = samples[:a]

        # randomly select solutions from previous Parate Set
        # This is to, first, preserve the high-quality solutions found along the evolutionary search process
        # second, to maintain the diversity of the population for further exploration of the evolutionary search.
        Q = self.opt.get("X")  # no-dominated solutions
        b = min(int(self.pop_size / 2), len(Q))
        idx = self.random_state.choice(np.arange(len(Q)), size=b)
        pop[a:(a + b)] = Population.new(X=Q[idx])

        # randomly generated solutions will be used to fill the population
        c = self.pop_size - a - b
        if c > 0:
            pop[(a + b):(a + b + c)] = self.initialization.sampling(self.problem, c, random_state=self.random_state)

        return pop



__all__ = [
    "DMOEADA",
    "DMOEADB",
    "DMOEADDEA",
    "DMOEADDEB",
    "DNSGA2A",
    "DNSGA2B",
    "MOEAD",
    "MOEADDE",
    "NSGA2",
    "Algorithm"
]

from .core.algorithm import Algorithm
from .dmoo.dmoead import DMOEADA, DMOEADB
from .dmoo.dmoeadde import DMOEADDEA, DMOEADDEB
from .dmoo.dnsga2 import DNSGA2A, DNSGA2B
from .moo.moead import MOEAD
from .moo.moeadde import MOEADDE
from .moo.nsga2 import NSGA2



import time

import numpy as np
from pymoo.core.population import Population
from pymoo.operators.survival.rank_and_crowding import RankAndCrowding

from pydmoo.algorithms.base.moo.moeadde import MOEADDE


class DMOEADDE(MOEADDE):

    def __init__(self,
                 perc_detect_change=0.1,
                 eps=0.0,
                 **kwargs):

        super().__init__(**kwargs)
        self.perc_detect_change = perc_detect_change
        self.eps = eps

    def setup(self, problem, **kwargs):
        assert not problem.has_constraints(), f"{self.__class__.__name__} only works for unconstrained problems."
        return super().setup(problem, **kwargs)

    def _detect_change_sample_part_population(self):
        pop = self.pop
        X, F = pop.get("X", "F")

        # the number of solutions to sample from the population to detect the change
        n_samples = int(np.ceil(len(pop) * self.perc_detect_change))

        # choose randomly some individuals of the current population to test if there was a change
        I = self.random_state.choice(np.arange(len(pop)), size=n_samples)
        samples = self.evaluator.eval(self.problem, Population.new(X=X[I]))

        # calculate the differences between the old and newly evaluated pop
        delta = ((samples.get("F") - F[I]) ** 2).mean()

        # if there is an average deviation bigger than eps -> we have a change detected
        change_detected = delta > self.eps
        return change_detected

    def _next_static_dynamic(self):
        # for dynamic environment
        pop = self.pop

        if self.state is None:

            change_detected = self._detect_change_sample_part_population()

            if change_detected:

                start_time = time.time()

                pop = self._response_mechanism()

                # reevaluate because we know there was a change
                self.evaluator.eval(self.problem, pop)

                if len(pop) > self.pop_size:
                    # do a survival to recreate rank and crowding of all individuals
                    # Modified by DynOpt on Dec 21, 2025
                    # n_survive=len(pop) -> n_survive=self.pop_size
                    pop = RankAndCrowding().do(self.problem, pop, n_survive=self.pop_size, random_state=self.random_state)

                self.pop = pop

                self.data["response_duration"] = time.time() - start_time

        return pop

    def _response_mechanism(self):
        """Response mechanism."""
        raise NotImplementedError


class DMOEADDEA(DMOEADDE):

    def __init__(self,
                 perc_detect_change=0.1,
                 eps=0.0,
                 perc_diversity=0.3,
                 **kwargs):
        super().__init__(perc_detect_change=perc_detect_change,
                         eps=eps,
                         **kwargs)

        self.perc_diversity = perc_diversity

    def _response_mechanism(self):
        """Response mechanism."""
        pop = self.pop
        X = pop.get("X")

        # recreate the current population without being evaluated
        pop = Population.new(X=X)

        # find indices to be replaced (introduce diversity)
        I = np.where(self.random_state.random(len(pop)) < self.perc_diversity)[0]

        # replace with randomly sampled individuals
        pop[I] = self.initialization.sampling(self.problem, len(I), random_state=self.random_state)

        return pop


class DMOEADDEB(DMOEADDE):

    def __init__(self,
                 perc_detect_change=0.1,
                 eps=0.0,
                 perc_diversity=0.3,
                 **kwargs):
        super().__init__(perc_detect_change=perc_detect_change,
                         eps=eps,
                         **kwargs)

        self.perc_diversity = perc_diversity

    def _response_mechanism(self):
        """Response mechanism."""
        pop = self.pop
        X = pop.get("X")

        # recreate the current population without being evaluated
        pop = Population.new(X=X)

        # find indices to be replaced (introduce diversity)
        I = np.where(self.random_state.random(len(pop)) < self.perc_diversity)[0]

        # replace by mutations of existing solutions (this occurs inplace)
        self.mating.mutation(self.problem, pop[I])

        return pop



"""
Includes modified code from [pymoo](https://github.com/anyoptimization/pymoo).

Sources:
    - [dnsga2.py](https://github.com/anyoptimization/pymoo/blob/main/pymoo/algorithms/moo/dnsga2.py).

Licensed under the Apache License, Version 2.0. Original copyright and license terms are preserved.

The framework was adapted to enable efficient aggregation of change response strategies.
"""
import time

import numpy as np
from pymoo.core.population import Population

from pydmoo.algorithms.base.moo.nsga2 import NSGA2


class DNSGA2(NSGA2):

    def __init__(self,
                 perc_detect_change=0.1,
                 eps=0.0,
                 **kwargs):

        super().__init__(**kwargs)
        self.perc_detect_change = perc_detect_change
        self.eps = eps

    def setup(self, problem, **kwargs):
        assert not problem.has_constraints(), f"{self.__class__.__name__} only works for unconstrained problems."
        return super().setup(problem, **kwargs)

    def _detect_change_sample_part_population(self):
        pop = self.pop
        X, F = pop.get("X", "F")

        # the number of solutions to sample from the population to detect the change
        n_samples = int(np.ceil(len(pop) * self.perc_detect_change))

        # choose randomly some individuals of the current population to test if there was a change
        I = self.random_state.choice(np.arange(len(pop)), size=n_samples)
        samples = self.evaluator.eval(self.problem, Population.new(X=X[I]))

        # calculate the differences between the old and newly evaluated pop
        delta = ((samples.get("F") - F[I]) ** 2).mean()

        # if there is an average deviation bigger than eps -> we have a change detected
        change_detected = delta > self.eps
        return change_detected

    def _infill_static_dynamic(self):
        # for dynamic environment
        pop = self.pop

        change_detected = self._detect_change_sample_part_population()

        if change_detected:

            start_time = time.time()

            pop = self._response_mechanism()

            # reevaluate because we know there was a change
            self.evaluator.eval(self.problem, pop)

            # do a survival to recreate rank and crowding of all individuals
            # Modified by DynOpt on Dec 21, 2025
            # n_survive=len(pop) -> n_survive=self.pop_size
            pop = self.survival.do(self.problem, pop, n_survive=self.pop_size, random_state=self.random_state)

            self.pop = pop

            self.data["response_duration"] = time.time() - start_time

        return pop

    def _response_mechanism(self):
        """Response mechanism."""
        raise NotImplementedError


class DNSGA2A(DNSGA2):

    def __init__(self,
                 perc_detect_change=0.1,
                 eps=0.0,
                 perc_diversity=0.3,
                 **kwargs):
        super().__init__(perc_detect_change=perc_detect_change,
                         eps=eps,
                         **kwargs)

        self.perc_diversity = perc_diversity

    def _response_mechanism(self):
        """Response mechanism."""
        pop = self.pop
        X = pop.get("X")

        # recreate the current population without being evaluated
        pop = Population.new(X=X)

        # find indices to be replaced (introduce diversity)
        I = np.where(self.random_state.random(len(pop)) < self.perc_diversity)[0]

        # replace with randomly sampled individuals
        pop[I] = self.initialization.sampling(self.problem, len(I), random_state=self.random_state)

        return pop


class DNSGA2B(DNSGA2):

    def __init__(self,
                 perc_detect_change=0.1,
                 eps=0.0,
                 perc_diversity=0.3,
                 **kwargs):
        super().__init__(perc_detect_change=perc_detect_change,
                         eps=eps,
                         **kwargs)

        self.perc_diversity = perc_diversity

    def _response_mechanism(self):
        """Response mechanism."""
        pop = self.pop
        X = pop.get("X")

        # recreate the current population without being evaluated
        pop = Population.new(X=X)

        # find indices to be replaced (introduce diversity)
        I = np.where(self.random_state.random(len(pop)) < self.perc_diversity)[0]

        # replace by mutations of existing solutions (this occurs inplace)
        self.mating.mutation(self.problem, pop[I])

        return pop



import numpy as np
from pymoo.core.population import Population
from pymoo.operators.survival.rank_and_crowding import RankAndCrowding

from pydmoo.algorithms.modern.moeadde_ktmm import MOEADDEKTMM
from pydmoo.core.bounds import clip_and_randomize
from pydmoo.core.inverse import closed_form_solution
from pydmoo.core.sample_gaussian import univariate_gaussian_sample


class MOEADDEIMKT(MOEADDEKTMM):
    """Inverse Modeling with Knowledge Transfer.

    Inverse Modeling for Dynamic Multiobjective Optimization with Knowledge Transfer In objective Space.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.size_pool = 10
        self.denominator = 0.5

    def _response_mechanism(self):
        """Response mechanism."""
        pop = self.pop
        X = pop.get("X")

        # recreate the current population without being evaluated
        pop = Population.new(X=X)

        # sample self.pop_size individuals in decision space
        samples_old = self.sampling_new_pop()

        # select self.pop_size/2 individuals with better convergence and diversity
        samples = samples_old[:int(len(samples_old)/2)]

        # knowledge in objective space
        means_stds, mean, std = self._in_decision_or_objective_space_1d(samples, "objective_space")
        mean_new, std_new = self._select_means_stds(means_stds, mean, std)

        # sample self.pop_size individuals in objective space
        F = univariate_gaussian_sample(mean_new, std_new, self.pop_size, random_state=self.random_state)

        # TODO
        # inverse mapping
        # X = FB
        B = closed_form_solution(samples.get("X"), samples.get("F"))

        # X = FB
        X = np.dot(F, B)

        # bounds
        if self.problem.has_bounds():
            xl, xu = self.problem.bounds()
            X = clip_and_randomize(X, xl, xu, random_state=self.random_state)

        # merge
        pop = Population.merge(samples_old, Population.new(X=X))

        return pop

    def sampling_new_pop(self):
        X = self.pop.get("X")

        if not self.problem.has_constraints():

            last_X = self.data.get("last_X", [])
            if len(last_X) == 0:
                last_X = X
            self.data["last_X"] = X

            d = np.mean(X - last_X, axis=0)

            radius = max(np.linalg.norm(d) / self.problem.n_obj, 0.1)

            X = X + d + self.random_state.uniform(low=-radius, high=radius, size=X.shape)

        # bounds
        if self.problem.has_bounds():
            xl, xu = self.problem.bounds()
            X = clip_and_randomize(X, xl, xu, random_state=self.random_state)

        samples = Population.new(X=X)
        samples = self.evaluator.eval(self.problem, samples)

        # do a survival to recreate rank and crowding of all individuals
        samples = RankAndCrowding().do(self.problem, samples, n_survive=len(samples))
        return samples


class MOEADDEIMKT0(MOEADDEIMKT):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)


class MOEADDEIMKT1(MOEADDEIMKT):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.size_pool = 2
        self.denominator = 0.5


class MOEADDEIMKT2(MOEADDEIMKT):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.size_pool = 4
        self.denominator = 0.5


class MOEADDEIMKT3(MOEADDEIMKT):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.size_pool = 6
        self.denominator = 0.5


class MOEADDEIMKT4(MOEADDEIMKT):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.size_pool = 8
        self.denominator = 0.5


class MOEADDEIMKT5(MOEADDEIMKT):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.size_pool = 10
        self.denominator = 0.5


class MOEADDEIMKT6(MOEADDEIMKT):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.size_pool = 12
        self.denominator = 0.5


class MOEADDEIMKT7(MOEADDEIMKT):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.size_pool = 14
        self.denominator = 0.5


class MOEADDEIMKT8(MOEADDEIMKT):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.size_pool = 16
        self.denominator = 0.5


class MOEADDEIMKT9(MOEADDEIMKT):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.size_pool = 18
        self.denominator = 0.5


class MOEADDEIMKT10(MOEADDEIMKT):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.size_pool = 20
        self.denominator = 0.5
